import numpy as np
from .misc import xy2rp, map_matrix, toroidal_period, div0


class Dipole(object):
    """
    magnetic dipole class
    """

    def __init__(self, **kwargs):
        """
        Initialize empty class
        """
        self.num = 0  # total number of dipoles
        self.nfp = 1  # toroidal_periodicity
        self.momentq = 1  # q expotent
        self.sp_switch = False  # switch to indicate if using spherical coordinates
        self.xyz_switch = False  # switch to indicate if using spherical coordinates
        self.old = False  # old format or new
        # self.symmetry = 2 # 0: no symmetry; 1: periodicity; 2: stellarator symmetry (+periodicity)
        if "mm" in kwargs:  # spherical coord
            self.ox = kwargs["ox"]
            self.oy = kwargs["oy"]
            self.oz = kwargs["oz"]
            self.mm = kwargs["mm"]
            self.mt = kwargs["mt"]
            self.mp = kwargs["mp"]
            self.pho = kwargs["pho"]
            self.momentq = kwargs["momentq"]
            self.Ic = kwargs["Ic"]
            self.Lc = kwargs["Lc"]
            self.num = len(self.ox)
            self.sp_switch = True
            self.filename = kwargs.get("filename", "dipole")
            self.name = kwargs.get(
                "name", ["pm_{:010d}".format(i) for i in range(1, self.num + 1)]
            )
            self.symm = kwargs.get("symm", 2 * np.ones(self.num, dtype=int))
            self.rho = self.pho ** self.momentq
        elif "mx" in kwargs:  # cartesian coord
            self.ox = kwargs["ox"]
            self.oy = kwargs["oy"]
            self.oz = kwargs["oz"]
            self.mx = kwargs["mx"]
            self.my = kwargs["my"]
            self.mz = kwargs["mz"]
            self.mm = np.sqrt(self.mx ** 2 + self.my ** 2 + self.mz ** 2)
            self.num = len(self.ox)
            self.pho = np.ones(self.num)
            self.Ic = np.zeros(self.num, dtype=int)
            self.Lc = np.zeros(self.num, dtype=int)
            self.xyz_switch = True
            self.filename = kwargs.get("filename", "dipole")
            self.name = kwargs.get(
                "name", ["pm_{:010d}".format(i) for i in range(1, self.num + 1)]
            )
            self.symm = kwargs.get("symm", 2 * np.ones(self.num, dtype=int))
            self.rho = np.ones(self.num)
        return

    @classmethod
    def open(cls, filename, verbose=False, **kwargs):
        """Read FAMUS dipoles

        Args:
            filename (str): path to open the file.
            verbose (bool, optional): Whether to print additional info. Defaults to False.

        Returns:
            Dipole: Dipole class
        """
        import pandas as pd

        with open(filename, "r") as coilfile:
            coilfile.readline()
            line = coilfile.readline()
            line = line.replace(",", " ")
            line = line.split()
            num = int(line[0])
            try:
                momentq = int(line[1])
            except:
                print("Moment Q factor was not read. Default=1.")
                momentq = 1
        data = pd.read_csv(filename, skiprows=3, header=None)
        symm = np.array(data[1], dtype=int)[0:num]
        name = np.array(data[2])[0:num]  # type string
        ox = np.array(data[3], dtype=float)[0:num]
        oy = np.array(data[4], dtype=float)[0:num]
        oz = np.array(data[5], dtype=float)[0:num]
        Ic = np.array(data[6], dtype=int)[0:num]
        mm = np.array(data[7], dtype=float)[0:num]
        pho = np.array(data[8], dtype=float)[0:num]
        Lc = np.array(data[9], dtype=int)[0:num]
        mp = np.array(data[10], dtype=float)[0:num]
        mt = np.array(data[11], dtype=float)[0:num]
        if verbose:
            print("Read {:d} dipoles from {:}".format(len(ox), filename))
        return cls(
            symm=symm,
            name=name,
            ox=ox,
            oy=oy,
            oz=oz,
            Ic=Ic,
            mm=mm,
            Lc=Lc,
            mp=mp,
            mt=mt,
            pho=pho,
            momentq=momentq,
            filename=filename,
        )

    @classmethod
    def read_dipole_old(cls, filename, zeta=0.0, zeta1=np.pi * 2, **kwargs):
        """read diploes from FOCUS format (old)

        Args:
            filename (str): path to open the file.
            zeta (float, optional): Starting toroidal angle for clipping. Defaults to 0.0.
            zeta1 (float, optional): [Ending toroidal angle for clipping]. Defaults to np.pi*2.

        Returns:
            Dipole: Dipole class
        """
        symm = []
        name = []
        ox = []
        oy = []
        oz = []
        mm = []
        mp = []
        mt = []
        Ic = []
        Lc = []
        with open(filename, "r") as coilfile:
            coilfile.readline()
            Ncoils = int(coilfile.readline()[0])
            for icoil in range(Ncoils):
                coilfile.readline()
                coilfile.readline()
                linelist = coilfile.readline().split()
                if int(linelist[0]) == 3:  # central current and background Bz
                    coilfile.readline()
                    linelist = coilfile.readline().split()
                elif int(linelist[0]) == 2:  # dipoles
                    symm.append(int(linelist[1]))
                    name.append(linelist[2])
                    coilfile.readline()
                    linelist = coilfile.readline().split()
                    r, phi = xy2rp(float(linelist[1]), float(linelist[2]))
                    if phi >= zeta and phi <= zeta1:
                        Lc.append(int(linelist[0]))
                        ox.append(float(linelist[1]))
                        oy.append(float(linelist[2]))
                        oz.append(float(linelist[3]))
                        Ic.append(int(linelist[4]))
                        mm.append(float(linelist[5]))
                        mt.append(float(linelist[6]))
                        mp.append(float(linelist[7]))
                elif int(linelist[0]) == 1:  # Fourier coils
                    for i in range(11):
                        coilfile.readline()
                else:
                    raise ValueError(
                        "Invalid coiltype = {:d}.".format(int(linelist[0]))
                    )
        nc = len(ox)
        if nc == 0:
            print("Warning: no dipoles was read from " + filename)
            return
        print(
            "Read {:d} dipoles from {:}. Please manually set self.old=True.".format(
                nc, filename
            )
        )
        symm = np.array(symm)
        ox = np.array(ox)
        oy = np.array(oy)
        oz = np.array(oz)
        mm = np.array(mm)
        mt = np.array(mt)
        mp = np.array(mp)
        Lc = np.array(Lc)
        Ic = np.array(Ic)
        pho = np.ones_like(ox)
        momentq = 1
        return cls(
            symm=symm,
            name=name,
            filename=filename,
            ox=ox,
            oy=oy,
            oz=oz,
            Ic=Ic,
            mm=mm,
            Lc=Lc,
            mp=mp,
            mt=mt,
            pho=pho,
            momentq=momentq,
        )

    @classmethod
    def from_regcoil(
        cls,
        regcoilname,
        winding,
        ilambda=-1,
        symmetry="full",
        num_pol=128,
        num_tor=128,
        m0=None,
        half_shift=True,
    ):
        """Initialize from REGCOIL current potential

        Args:
            regcoilname (str): REGCOIL netcdf output file.
            winding (str, optional): NESCOIL input format winding surface.
            ilambda (int, optional): Lambda index in REGCOIL output. Defaults to -1.
            symmetry (str, optional): Stellarator symmetry option. Defaults to 'full'.
            num_pol (int, optional): Number of poloidal dipoles. Defaults to 128.
            num_tor (int, optional): Number of toroidal dipoles. Defaults to 128.
            m0 ([type], optional): Magnetization limit per thickness. Defaults to None.
            half_shift (bool, optional): Logical flag to determine if half-grid shifted. Defaults to True.

        Returns:
            None
        """
        # read regcoil output
        from scipy.io import netcdf

        f = netcdf.netcdf_file(regcoilname, "r", mmap=False)
        nfp = f.variables["nfp"][()]
        xm_potential = f.variables["xm_potential"][()]
        xn_potential = f.variables["xn_potential"][()]
        phi_mn = f.variables["single_valued_current_potential_mn"][()][ilambda, :]
        symmetry_option = f.variables["symmetry_option"][()]
        f.close()

        mn_max = len(xm_potential)
        phi_sin = np.zeros(mn_max)
        phi_cos = np.zeros(mn_max)
        if symmetry_option == 1:
            phi_sin = phi_mn[0:mn_max]
        elif symmetry_option == 2:
            phi_cos = phi_mn[0:mn_max]
        elif symmetry_option == 3:
            phi_sin = phi_mn[0:mn_max]
            phi_cos = phi_mn[mn_max:]
        else:
            raise ValueError(
                "Something wrong the symmetry_option: {:}".format(symmetry_option)
            )

        def func_new(theta, zeta):
            """get current potential of (theta, zeta)

            Parameters:
            theta -- float array_like, poloidal angle
            zeta -- float array_like, toroidal angle value

            Returns:
            phi -- float array_like
            """
            assert len(np.atleast_1d(theta)) == len(
                np.atleast_1d(zeta)
            ), "theta, zeta should be equal size"
            # mt - nz (in matrix)
            _mtnz = np.matmul(
                np.reshape(xm_potential, (-1, 1)), np.reshape(theta, (1, -1))
            ) - np.matmul(np.reshape(xn_potential, (-1, 1)), np.reshape(zeta, (1, -1)))
            _sin = np.sin(_mtnz)
            _cos = np.cos(_mtnz)

            phi = np.matmul(np.reshape(phi_sin, (1, -1)), _sin) + np.matmul(
                np.reshape(phi_cos, (1, -1)), _cos
            )
            return phi.ravel()

        # initialize dipoles
        num = num_pol * num_tor
        if symmetry.lower() == "full":
            zeta_end = 2 * np.pi
            symm = 0 + np.zeros(num, dtype=np.int)
        elif symmetry.lower() == "one":
            zeta_end = 2 * np.pi / nfp
            symm = 1 + np.zeros(num, dtype=np.int)
        elif symmetry.lower() == "half":
            zeta_end = 2 * np.pi / nfp / 2
            symm = 2 + np.zeros(num, dtype=np.int)
        else:
            raise ValueError("No such symmetry option!")

        ox = []
        oy = []
        oz = []
        mt = []
        mp = []
        mm = []

        # discretization
        if half_shift:  # toroidally shift half grid
            half = zeta_end / (num_tor * 2)
        else:
            half = 0
        theta_array = np.linspace(0, 2 * np.pi, num_pol, endpoint=False)
        zeta_array = half + np.linspace(0, zeta_end, num_tor, endpoint=False)
        dtheta = 2 * np.pi / num_pol
        dzeta = zeta_end / num_tor
        tv, zv = np.meshgrid(theta_array, zeta_array, indexing="ij")

        from .surface import FourSurf

        wind = FourSurf.read_winding_surfce(winding)
        data = wind.xyz(tv, zv, normal=True)
        phi = func_new(tv, zv)
        ox = data[0]
        oy = data[1]
        oz = data[2]
        n = data[3]
        nn = np.linalg.norm(n, axis=1)
        mt = np.arccos(n[:, 2] / nn)
        mp = np.arctan2(n[:, 1], n[:, 0])
        if m0 is None:  # not specify the allowable magnetization
            mm = -phi * nn * dtheta * dzeta
            pho = np.ones_like(mm)
        else:
            mm = m0 * nn * dtheta * dzeta
            pho = -phi / m0

        return cls(
            symm=symm,
            ox=ox,
            oy=oy,
            oz=oz,
            mp=mp,
            mt=mt,
            mm=mm,
            pho=pho,
            Ic=np.ones_like(ox, dtype=int),
            Lc=np.ones_like(ox, dtype=int),
            momentq=1,
        )

    def sp2xyz(self):
        """
        spherical coordinates to cartesian coordinates
        """
        assert self.sp_switch == True, "You are not using spherical coordinates"
        if self.old:
            self.mx = self.mm * np.sin(self.mt) * np.cos(self.mp)
            self.my = self.mm * np.sin(self.mt) * np.sin(self.mp)
            self.mz = self.mm * np.cos(self.mt)
        else:
            self.mx = (
                self.mm * self.pho ** self.momentq * np.sin(self.mt) * np.cos(self.mp)
            )
            self.my = (
                self.mm * self.pho ** self.momentq * np.sin(self.mt) * np.sin(self.mp)
            )
            self.mz = self.mm * self.pho ** self.momentq * np.cos(self.mt)
        self.xyz_switch = True
        return

    def xyz2sp(self):
        """
        cartesian coordinates to spherical coordinates
        """
        assert self.xyz_switch is True, "You are not using cartesian coordinates"
        if self.old:
            self.mm = np.sqrt(self.mx * self.mx + self.my * self.my + self.mz * self.mz)
            div = np.divide(
                self.mz, self.mm, out=np.zeros_like(self.mz), where=self.mm != 0
            )
            self.mt = np.arccos(div0(self.mz, self.mm))
            # self.pho = np.ones_like(self.mm)
            # self.momentq = 1
        else:
            mmag = np.sqrt(self.mx * self.mx + self.my * self.my + self.mz * self.mz)
            self.pho = np.power(div0(mmag, self.mm), 1.0 / self.momentq)
            self.mt = np.arccos(div0(self.mz, (self.mm * self.pho ** self.momentq)))
        self.mp = np.arctan2(self.my, self.mx)
        self.sp_switch = True
        return

    def save(self, filename, unique=False, tol=0):
        """write diploes from FOCUS format

        Args:
            filename (str): FOCUS file name.
            unique (bool, optional): Writing dipole every self.nfp term. Defaults to False.
            tol (float, optional): tolerance to skip zeros. Defaults to 0.
        """
        if not self.sp_switch:
            self.xyz2sp()
        cond = np.abs(self.rho) >= tol
        with open(filename, "w") as wfile:
            wfile.write(" # Total number of dipoles,  momentq \n")
            if unique:
                wfile.write(
                    "{:6d},  {:4d}\n".format(
                        np.count_nonzero(cond) / self.nfp, self.momentq
                    )
                )
            else:
                wfile.write(
                    "{:6d},  {:4d}\n".format(np.count_nonzero(cond), self.momentq)
                )
            if self.old:
                for icoil in range(self.num):
                    if unique:
                        if np.mod(icoil, self.nfp) == 0:
                            continue
                    if not cond[icoil]:
                        continue
                    wfile.write(
                        "#-----------------{}---------------------------\n".format(
                            icoil + 1
                        )
                    )
                    wfile.write("#coil_type   symm  coil_name \n")
                    wfile.write(
                        "   {:1d}  {:1d}  {:} \n".format(
                            2, self.symm[icoil], self.name[icoil]
                        )
                    )
                    wfile.write("#  Lc  ox   oy   oz  Ic  I  mt  mp \n")
                    wfile.write(
                        "{:6d} {:23.15E} {:23.15E} {:23.15E} {:6d} {:23.15E} {:23.15E} {:23.15E}\n".format(
                            self.Lc[icoil],
                            self.ox[icoil],
                            self.oy[icoil],
                            self.oz[icoil],
                            self.Ic[icoil],
                            self.mm[icoil],
                            self.mt[icoil],
                            self.mp[icoil],
                        )
                    )
            else:
                wfile.write(
                    "#coiltype, symmetry,  coilname,  ox,  oy,  oz,  Ic,  M_0,  pho,  Lc,  mp,  mt \n"
                )
                for i in range(self.num):
                    if unique:
                        if np.mod(i, self.nfp) == 0:
                            continue
                    if not cond[i]:
                        continue
                    wfile.write(
                        " 2, {:1d}, {:}, {:15.8E}, {:15.8E}, {:15.8E}, {:2d}, {:15.8E},"
                        "{:15.8E}, {:2d}, {:15.8E}, {:15.8E} \n".format(
                            self.symm[i],
                            self.name[i],
                            self.ox[i],
                            self.oy[i],
                            self.oz[i],
                            self.Ic[i],
                            self.mm[i],
                            self.pho[i],
                            self.Lc[i],
                            self.mp[i],
                            self.mt[i],
                        )
                    )
        return

    def toVTK(
        self, vtkname, dim=(1), close=False, ntnz=False, toroidal=False, **kwargs
    ):
        """write dipole data into a VTK file

        Args:
            vtkname (str): VTK filename, will be appended with .vts or .vtu.
            dim (tuple, optional): Dimension information if saved as structured grids. Defaults to (1).
            close (bool, optional): Logical flag to manually close the gaps. Defaults to False.
            ntnz (bool, optional): Logical flag of theta-zeta order. Defaults to False.
            toroidal (bool, optional): Logical flag of filling in the toroidal gap. Defaults to False.

        Returns:
            None
        """
        from pyevtk.hl import gridToVTK, pointsToVTK

        if not self.xyz_switch:
            self.sp2xyz()
        if vtkname is None:
            vtkname = self.filename
        dim = np.atleast_1d(dim)
        if len(dim) == 1:  # save as points
            print("write VTK as points")
            data = {"m": (self.mx, self.my, self.mz)}
            if not self.old:
                data.update({"rho": self.pho ** self.momentq})
            data.update(kwargs)
            pointsToVTK(
                vtkname, self.ox, self.oy, self.oz, data=data
            )  # .update(kwargs))
        else:  # save as surfaces
            assert len(dim) == 3
            print("write VTK as closed surface")
            if close:
                # manually close the gap
                phi = 2 * np.pi / self.nfp

                def map_toroidal(vec):
                    rotate = np.array(
                        [
                            [np.cos(phi), np.sin(phi), 0],
                            [-np.sin(phi), np.cos(phi), 0],
                            [0, 0, 1],
                        ]
                    )
                    return np.matmul(vec, rotate)

                data_array = {
                    "ox": self.ox,
                    "oy": self.oy,
                    "oz": self.oz,
                    "mx": self.mx,
                    "my": self.my,
                    "mz": self.mz,
                    "Ic": self.Ic,
                    "rho": self.pho ** self.momentq,
                }
                data_array.update(kwargs)
                nr, nz, nt = dim
                for key in list(data_array.keys()):
                    new_vec = np.zeros((nr, nz + 1, nt + 1))
                    for ir in range(nr):
                        new_vec[ir, :, :] = map_matrix(
                            np.reshape(data_array[key], dim)[ir, :, :]
                        )
                    if toroidal:
                        data_array[key] = new_vec
                    else:
                        if ntnz:
                            data_array[key] = np.ascontiguousarray(new_vec[:, :, :-1])
                        else:
                            data_array[key] = np.ascontiguousarray(new_vec[:, :-1, :])
                ox = np.copy(data_array["ox"])
                oy = np.copy(data_array["oy"])
                oz = np.copy(data_array["oz"])
                del data_array["ox"]
                del data_array["oy"]
                del data_array["oz"]
                data_array["m"] = (data_array["mx"], data_array["my"], data_array["mz"])
                if toroidal and self.nfp >= 1:  # not quite sure if should include nfp=1
                    for ir in range(nr):
                        if ntnz:
                            xyz = map_toroidal(
                                np.transpose([ox[ir, :, 0], oy[ir, :, 0], oz[ir, :, 0]])
                            )
                            ox[ir, :, nz] = xyz[:, 0]
                            oy[ir, :, nz] = xyz[:, 1]
                            oz[ir, :, nz] = xyz[:, 2]
                            moment = map_toroidal(
                                np.transpose(
                                    [
                                        data_array["mx"][ir, :, 0],
                                        data_array["my"][ir, :, 0],
                                        data_array["mz"][ir, :, 0],
                                    ]
                                )
                            )
                            data_array["m"][0][ir, :, nz] = moment[:, 0]
                            data_array["m"][1][ir, :, nz] = moment[:, 1]
                            data_array["m"][2][ir, :, nz] = moment[:, 2]
                        else:
                            xyz = map_toroidal(
                                np.transpose([ox[ir, 0, :], oy[ir, 0, :], oz[ir, 0, :]])
                            )
                            ox[ir, nz, :] = xyz[:, 0]
                            oy[ir, nz, :] = xyz[:, 1]
                            oz[ir, nz, :] = xyz[:, 2]
                            moment = map_toroidal(
                                np.transpose(
                                    [
                                        data_array["mx"][ir, 0, :],
                                        data_array["my"][ir, 0, :],
                                        data_array["mz"][ir, 0, :],
                                    ]
                                )
                            )
                            data_array["m"][0][ir, nz, :] = moment[:, 0]
                            data_array["m"][1][ir, nz, :] = moment[:, 1]
                            data_array["m"][2][ir, nz, :] = moment[:, 2]
                del data_array["mx"]
                del data_array["my"]
                del data_array["mz"]
                gridToVTK(vtkname, ox, oy, oz, pointData=data_array)
                return
            else:
                ox = np.reshape(self.ox[: self.num], dim)
                oy = np.reshape(self.oy[: self.num], dim)
                oz = np.reshape(self.oz[: self.num], dim)
                mx = np.reshape(self.mx[: self.num], dim)
                my = np.reshape(self.my[: self.num], dim)
                mz = np.reshape(self.mz[: self.num], dim)
                rho = np.reshape(self.pho[: self.num] ** self.momentq, dim)
                Ic = np.reshape(self.Ic[: self.num], dim)
            data = {"m": (mx, my, mz), "rho": rho, "Ic": Ic}
            data.update(kwargs)
            gridToVTK(vtkname, ox, oy, oz, pointData=data)
        return

    def full_period(self, nfp=1, dim=None):
        """
        map from one period to full periods
        """
        assert nfp >= 1
        self.nfp = nfp
        if not self.xyz_switch:
            self.sp2xyz()
        # change order
        if dim is not None:
            self.ox = np.ravel(
                np.transpose(np.reshape(self.ox, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.oy = np.ravel(
                np.transpose(np.reshape(self.oy, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.oz = np.ravel(
                np.transpose(np.reshape(self.oz, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.mx = np.ravel(
                np.transpose(np.reshape(self.mx, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.my = np.ravel(
                np.transpose(np.reshape(self.my, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.mz = np.ravel(
                np.transpose(np.reshape(self.mz, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.mm = np.ravel(
                np.transpose(np.reshape(self.mm, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.Ic = np.ravel(
                np.transpose(np.reshape(self.Ic, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.Lc = np.ravel(
                np.transpose(np.reshape(self.Lc, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
            self.pho = np.ravel(
                np.transpose(np.reshape(self.pho, dim)[::-1, ::-1, ::-1], (2, 0, 1))
            )
        if np.mean(self.symm) == 2:
            # get the stellarator symmetry part first
            # Here, we assume no dipoles on the symmetry plane, or only half are listed.
            self.num *= 2
            self.name = np.concatenate((self.name, self.name[::-1]))
            self.ox = np.concatenate((self.ox, self.ox[::-1]))
            self.oy = np.concatenate((self.oy, self.oy[::-1] * (-1)))
            self.oz = np.concatenate((self.oz, self.oz[::-1] * (-1)))
            self.mx = np.concatenate((self.mx, self.mx[::-1] * (-1)))
            self.my = np.concatenate((self.my, self.my[::-1]))
            self.mz = np.concatenate((self.mz, self.mz[::-1]))
            self.mm = np.concatenate((self.mm, self.mm[::-1]))
            self.pho = np.concatenate((self.pho, self.pho[::-1]))
            self.Ic = np.concatenate((self.Ic, self.Ic[::-1]))
            self.Lc = np.concatenate((self.Lc, self.Lc[::-1]))
        xyz = toroidal_period(np.transpose([self.ox, self.oy, self.oz]), self.nfp)
        self.ox = xyz[:, 0].copy()
        self.oy = xyz[:, 1].copy()
        self.oz = xyz[:, 2].copy()
        moment = toroidal_period(np.transpose([self.mx, self.my, self.mz]), self.nfp)
        self.mx = moment[:, 0].copy()
        self.my = moment[:, 1].copy()
        self.mz = moment[:, 2].copy()
        self.mm = np.tile(self.mm, self.nfp)
        self.pho = np.tile(self.pho, self.nfp)
        self.rho = self.pho ** self.momentq
        self.Ic = np.tile(self.Ic, self.nfp)
        self.Lc = np.tile(self.Lc, self.nfp)
        self.name = np.tile(self.name, self.nfp)
        self.num *= self.nfp
        self.symm = np.zeros(self.num, dtype=int)
        self.sp_switch = False
        return

    def inverse(self):
        """get the stellarator symmetric part?"""
        if not self.xyz_switch:
            self.sp2xyz()
        self.ox = np.copy(self.ox[::-1])
        self.oy = np.copy(self.oy[::-1] * (-1))
        self.oz = np.copy(self.oz[::-1] * (-1))
        self.mx = np.copy(self.mx[::-1] * (-1))
        self.my = np.copy(self.my[::-1])
        self.mz = np.copy(self.mz[::-1])
        self.mm = np.copy(self.mm[::-1])
        self.pho = np.copy(self.pho[::-1])
        self.Ic = np.copy(self.Ic[::-1])
        self.Lc = np.copy(self.Lc[::-1])
        return

    def change_momentq(self, newq):
        """
        change the q factor for the normalized density
        """
        assert newq > 0
        pho = self.pho ** self.momentq
        # get signs
        self.sp2xyz()
        sign = np.sign(pho)
        if newq % 2 == 0:  # even number, flip the orientation when negative
            # self.mx *= sign
            # self.my *= sign
            # self.mz *= sign
            self.momentq = newq
            self.xyz2sp()
            # convert to positive rho
            # self.pho = np.power(np.abs(pho), 1.0/newq)
        else:  # odd exponetial index
            self.momentq = newq
            # convert to positive rho
            self.pho = np.power(np.abs(pho), 1.0 / newq)
            self.pho *= sign
        return

    def plot_rho_profile(self, lower=0, upper=1, nrange=10, nofigure=False, **kwargs):
        import matplotlib.pyplot as plt

        pho = np.abs(self.pho ** self.momentq)
        zone = np.linspace(lower, upper, nrange + 1, endpoint=True)
        count = []
        for i in range(nrange - 1):
            count.append(((pho >= zone[i]) & (pho < zone[i + 1])).sum())
        count.append(((pho >= zone[nrange - 1]) & (pho <= zone[nrange])).sum())
        count = np.array(count)
        if not nofigure:
            if plt.get_fignums():
                fig = plt.gcf()
                ax = plt.gca()
            else:
                fig, ax = plt.subplots()
            plt.bar(
                zone[:-1], 100 * count / float(self.num), width=0.9 / nrange, **kwargs
            )
            ax.set_xlabel(r"$|\rho|$", fontsize=15)
            ax.set_ylabel("fraction [%]", fontsize=15)
            plt.xticks(fontsize=14)
            plt.yticks(fontsize=14)
        return count

    def volume(self, magnitization=1.1e6, **kwargs):
        self.total_moment = np.sum(np.abs(self.rho * self.mm))
        return self.total_moment / magnitization

    def orientation(self, unit=True, uniform=False):
        oldq = self.momentq
        self.change_momentq(2)  # to a even number
        self.change_momentq(oldq)  # recover
        if unit:
            self.mm[:] = 1.0
        if uniform:
            self.pho[:] = 1.0
        self.sp2xyz()
        return

    def round_angle(self):
        """round off the moment orientation to the closest preferred axis

        Returns:
            Dipole class: return a new dipole class with rounded orientations
        """
        from copy import deepcopy

        new = deepcopy(self)
        phi1 = np.arctan2(new.oy, new.ox)
        phi2 = phi1 + np.pi / 2
        mxyz = np.transpose(
            [
                np.sin(new.mt) * np.cos(new.mp),
                np.sin(new.mt) * np.sin(new.mp),
                np.cos(new.mt),
            ]
        )
        xvec = np.transpose([np.cos(phi1), np.sin(phi1), np.zeros_like(phi1)])
        yvec = np.transpose([np.cos(phi2), np.sin(phi2), np.zeros_like(phi1)])
        zvec = np.reshape(np.tile([0, 0, 1], new.num), (-1, 3))
        cosx = np.sum(mxyz * xvec, axis=1)
        cosy = np.sum(mxyz * yvec, axis=1)
        cosz = np.sum(mxyz * zvec, axis=1)
        cos_arr = np.transpose([cosx[:], cosy[:], cosz[:]])
        argmin = np.argmax(np.abs(cos_arr), axis=1)
        theta_arr = np.reshape(np.tile([np.pi / 2, np.pi / 2, 0], new.num), (-1, 3))
        phi_arr = np.transpose([phi1, phi2, np.zeros_like(phi1)])
        new.mt = np.array([theta_arr[i, argmin[i]] for i in range(new.num)])
        new.mp = np.array([phi_arr[i, argmin[i]] for i in range(new.num)])
        sign_arr = np.sign([cos_arr[i, argmin[i]] for i in range(new.num)])
        new.pho *= sign_arr
        new.rho *= sign_arr
        return new

    def mimic(self, template, **kwargs):
        """Mimic rho, mp, mt following the "template" dipole set

        Args:
            template (Dipole class): The template Dipole class to be mirrored.

        Returns:
            None: None
        """
        # find the cloest dipole in real space
        def closest(i):
            dis = (
                (self.ox[i] - template.ox) ** 2
                + (self.oy[i] - template.oy) ** 2
                + (self.oz[i] - template.oz) ** 2
            )
            return np.argmin(dis)

        # assign rho, mt, mp
        for i in range(self.num):
            ind = closest(i)
            self.rho[i] = template.rho[ind]
            self.mt[i] = template.mt[ind]
            self.mp[i] = template.mp[ind]
        self.pho = self.rho ** (1.0 / self.momentq)
        return

    def plot(self, engine="pyplot", start=0, end=None, **kwargs):
        if end is None:
            end = self.num
        if engine == "pyplot":
            import matplotlib.pyplot as plt

            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            ax.scatter(
                self.ox[start:end], self.oy[start:end], self.oz[start:end], **kwargs
            )
        elif engine == "plotly":
            import plotly.express as px

            fig = px.scatter_3d(
                x=self.ox[start:end],
                y=self.oy[start:end],
                z=self.oz[start:end],
                color=self.rho[start:end],
                **kwargs
            )
            fig.show()
        return

    def bfield(self, pos):
        """Calculate the magnetic field at an arbitrary position.
           No symmetry info considered for now.

        Args:
            pos (array_like): [x,y,z] Cartesian coordinates in space.

        Returns:
            numpy.array: The total magnetic field produced by all dipoles
        """
        # calculate mx, my, mz if needed
        if not self.xyz_switch:
            self.sp2xyz()
        # Biot-Savart law
        pos = np.reshape(pos, (3, 1))
        oxyz = np.asarray([self.ox, self.oy, self.oz])
        rxyz = oxyz - pos
        r = np.linalg.norm(rxyz, axis=0)
        mxyz = np.asarray([self.mx, self.my, self.mz])
        Bvec = 3 * np.sum(mxyz * rxyz, axis=0) / r ** 5 * rxyz - 1 / r ** 3 * mxyz
        return 1e-7 * np.sum(Bvec, axis=1)

    def __repr__(self):
        return "FAMUS dipole class, num={:d}, symmetry={:}, filename={:}".format(
            self.num, np.mean(self.symm), self.filename
        )

    def __add__(self, other):
        """Combine two dipole files together"""
        if not self.sp_switch:
            self.xyz2sp()
        assert self.momentq == other.momentq, "Two classes should use the same momentq."
        return Dipole(
            symm=np.concatenate((self.symm, other.symm)),
            name=np.concatenate((self.name, other.name)),
            ox=np.concatenate((self.ox, other.ox)),
            oy=np.concatenate((self.oy, other.oy)),
            oz=np.concatenate((self.oz, other.oz)),
            Ic=np.concatenate((self.Ic, other.Ic)),
            mm=np.concatenate((self.mm, other.mm)),
            Lc=np.concatenate((self.Lc, other.Lc)),
            mp=np.concatenate((self.mp, other.mp)),
            mt=np.concatenate((self.mt, other.mt)),
            pho=np.concatenate((self.pho, other.pho)),
            momentq=self.momentq,
            filename=self.filename + "+" + other.filename,
        )

    def truncate(self, cond):
        """Truncate dipole set following a condition

        Args:
            cond (array-like, bool): Slicing conditions

        Returns:
            Dipole: Truncated dipole
        """
        return Dipole(
            symm=self.symm[cond],
            name=self.name,
            ox=self.ox[cond],
            oy=self.oy[cond],
            oz=self.oz[cond],
            Ic=self.Ic[cond],
            mm=self.mm[cond],
            Lc=self.Lc[cond],
            mp=self.mp[cond],
            mt=self.mt[cond],
            pho=self.pho[cond],
            momentq=self.momentq,
            filename=self.filename + "_clip",
        )


class GAdipole(Dipole):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


if __name__ == "main":
    pass


"""
Some useful functions used for the PM4STELL project
"""
import numpy as np
import pandas as pd
from .dipole import Dipole


def blocks2vtk(
    block_file, vtk_file, moment_file=None, dipole_file=None, clip=0, **kwargs
):
    """Write a VTK file from the blocks file

    Args:
        block_file (str): File name and path to the `blocks` file.
        vtk_file (str): VTK file name to be saved.
        moment_file (str, optional): File name and path to the `moments` file. Defaults to None.
        dipole_file (str, optional): File name and path to the FAMUS dipole file (`*.focus`). Defaults to None.
        clip (int, optional): The threshold value to clip magents with rho>=clip. Defaults to 0.

    Returns:
        meshio.Mesh: The constructed `meshio.Mesh` object.
    """
    import meshio

    assert ".vtk" in vtk_file, ".vtk must be in the filename."
    assert clip >= 0, "the clip value should be >=0."
    blocks = pd.read_csv(block_file, skiprows=1)
    # remove space in headers
    blocks.rename(columns=lambda x: x.strip(), inplace=True)
    nmag = len(
        blocks["xb1"],
    )
    cond = np.full((nmag), True)
    # parse moment data
    kwargs.setdefault("cell_data", {})
    if moment_file is not None:
        moments = pd.read_csv(moment_file, skiprows=1)
        moments.rename(columns=lambda x: x.strip(), inplace=True)
        kwargs["cell_data"].setdefault(
            "m",
            [
                np.ascontiguousarray(
                    np.transpose([moments["Mx"], moments["My"], moments["Mz"]])
                )
            ],
        )
        kwargs["cell_data"].setdefault("rho", [moments["rho"]])
        kwargs["cell_data"].setdefault("type", [moments["type"]])
    if dipole_file is not None:
        dipoles = Dipole.open(dipole_file)
        dipoles.sp2xyz()
        cond = dipoles.rho >= clip
        kwargs["cell_data"].setdefault(
            "m",
            [
                np.ascontiguousarray(
                    np.transpose([dipoles.mx[cond], dipoles.my[cond], dipoles.mz[cond]])
                )
            ],
        )
        kwargs["cell_data"].setdefault("rho", [dipoles.rho[cond]])
        kwargs["cell_data"].setdefault("Lc", [dipoles.Lc[cond]])
    # update blocks based on cond
    blocks = blocks.loc[cond]
    x = np.concatenate(
        [
            blocks["xb1"],
            blocks["xb2"],
            blocks["xb3"],
            blocks["xb4"],
            blocks["xt1"],
            blocks["xt2"],
            blocks["xt3"],
            blocks["xt4"],
        ]
    )
    y = np.concatenate(
        [
            blocks["yb1"],
            blocks["yb2"],
            blocks["yb3"],
            blocks["yb4"],
            blocks["yt1"],
            blocks["yt2"],
            blocks["yt3"],
            blocks["yt4"],
        ]
    )
    z = np.concatenate(
        [
            blocks["zb1"],
            blocks["zb2"],
            blocks["zb3"],
            blocks["zb4"],
            blocks["zt1"],
            blocks["zt2"],
            blocks["zt3"],
            blocks["zt4"],
        ]
    )
    # write VTK
    nmag = np.count_nonzero(cond)
    ind = np.reshape(np.arange(len(x)), (8, nmag))
    points = np.ascontiguousarray(np.transpose([x, y, z]))
    hedrs = [ind[:, i] for i in range(nmag)]
    data = meshio.Mesh(points=points, cells=[("hexahedron", hedrs)], **kwargs)
    data.write(vtk_file)
    return data


def blocks2ficus(
    block_file,
    ficus_file,
    moment_file=None,
    dipole_file=None,
    magnitization=1.1e6,
    clip=None,
    **kwargs
):
    import pandas as pd
    import FICUS.Magnet3D as m3

    blocks = pd.read_csv(block_file, skiprows=1)
    blocks.to_csv("tmp.csv", columns=blocks.columns[7:], index=False)
    corner = m3.Magnet_3D("tmp.csv")
    muse_data = corner.export_source()
    muse_data[:, -1] = magnitization
    # read dipole moment
    if moment_file is not None:
        moments = pd.read_csv(moment_file, skiprows=1)
        moments.rename(columns=lambda x: x.strip(), inplace=True)
        mx = moments["Mx"].to_numpy()
        my = moments["My"].to_numpy()
        mz = moments["Mz"].to_numpy()
        rho = moments["rho"].to_numpy()
    if dipole_file is not None:
        dipoles = Dipole.open(dipole_file)
        dipoles.sp2xyz()
        mx = dipoles.mx
        my = dipoles.my
        mz = dipoles.mz
        rho = dipoles.rho
    # filter
    if clip is None:
        cond = np.full(np.shape(rho), True)
    else:
        cond = rho > clip
    muse_data = np.concatenate(
        [muse_data, mx[:, np.newaxis], my[:, np.newaxis], mz[:, np.newaxis]], axis=1
    )
    dt = pd.DataFrame(
        muse_data[cond],
        columns=[
            "ox",
            "oy",
            "oz",
            "nx",
            "ny",
            "nz",
            "ux",
            "uy",
            "uz",
            "H",
            "L",
            "M",
            "mx",
            "my",
            "mz",
        ],
    )
    dt.to_csv(ficus_file, index=False)
    return dt


def read_ansys_bfield(filename):
    ansys = pd.read_csv(filename, skiprows=[0], delim_whitespace=True, header=None)
    ansys.columns = ["x", "y", "z", "Bx", "By", "Bz"]
    return ansys


import setuptools
from numpy.distutils.core import Extension, setup
from numpy.distutils.fcompiler import get_default_fcompiler

# from coilpy import __version__

__version__ = "0.3.13"

with open("README.md", "r") as fh:
    long_description = fh.read()

# fortran compiler
compiler = get_default_fcompiler()
# set some fortran compiler-dependent flags
f90flags = []
if compiler == "gnu95":
    f90flags.append("-ffree-line-length-none")
elif compiler == "intel" or compiler == "intelem":
    pass
f90flags.append("-O3")

ext = Extension(
    name="coilpy_fortran",
    sources=[
        "coilpy/fortran/biotsavart.f90",
    ],
    extra_f90_compile_args=f90flags,
)

setup(
    name="coilpy",
    version=__version__,
    description="Plotting and data processing tools for plasma and coil",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    url="https://github.com/zhucaoxiang/CoilPy",
    author="Caoxiang Zhu",
    author_email="caoxiangzhu@gmail.com",
    license="GNU 3.0",
    packages=setuptools.find_packages(),
    ext_modules=[ext],
)



# modified from stellopt.pySTEL

import numpy as np
import xarray
import matplotlib.pyplot as plt
from .misc import trig2real
from .surface import FourSurf

__all__ = ["VMECout"]


class VMECout(object):
    """
    Class to parse a VMEC wout file

    :param filename: filename passed to xarray.open_dataset()

    The entire dataset is stored in `self.wout` and you can access
    to variables via `self.wout['iotaf'].values`.
    The flux surfaces are all parsed as `FourSurf` classes and stored
    in the list of `self.surface`.
    Magnetic fields are Fourier transformed to real space
    (only the stellarator symmetric part, `bmnc`) and store in `self.data['b]`.

    `self.plot` has several options to plot the profiles, etc.

    """

    def __init__(self, filename, **kwargs):
        self.wout = xarray.open_dataset(filename)
        self.data = {}
        self.data["ns"] = int(self.wout["ns"].values)
        self.data["nfp"] = int(self.wout["nfp"].values)
        self.data["nu"] = int(self.wout["mpol"].values * 4)
        self.data["nv"] = int(self.wout["ntor"].values * 4)
        self.data["nflux"] = np.linspace(
            0, 1, self.data["ns"]
        )  # np.ndarray((self.data['ns'],1))
        self.data["theta"] = np.linspace(
            0, 2 * np.pi, self.data["nu"]
        )  # np.ndarray((self.data['nu'],1))
        self.data["zeta"] = np.linspace(
            0, 2 * np.pi, self.data["nv"]
        )  # np.ndarray((self.data['nv'],1))
        self.surface = []
        self.data["b"] = []
        for i in range(self.data["ns"]):
            self.surface.append(
                FourSurf(
                    xm=self.wout["xm"].values,
                    xn=self.wout["xn"].values,
                    rbc=self.wout["rmnc"][i].values,
                    rbs=np.zeros_like(self.wout["rmnc"][i].values),
                    zbs=self.wout["zmns"][i].values,
                    zbc=np.zeros_like(self.wout["zmns"][i].values),
                )
            )
            self.data["b"].append(
                trig2real(
                    self.data["theta"],
                    self.data["zeta"],
                    self.wout["xm_nyq"].values,
                    self.wout["xn_nyq"].values / self.data["nfp"],
                    self.wout["bmnc"][i].values,
                )
            )
        return

    def plot(self, plot_name="none", ax=None, **kwargs):
        """Plot various VMEC quantities

        Args:
            plot_name (str, optional): The quantity to be plotted, should be one of
                                       iota, q, pressue, <Buco>, <Bvco>, <jcuru>, <jcurv>,
                                       <j.B>, LPK, none. Defaults to 'none'.
            ax (Matplotlib axis, optional): The Matplotlib axis to be plotted on. Defaults to None.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        if plot_name == "none":
            print("You can plot: iota, q, pressue, <Buco>, <Bvco>, <jcuru>, <jcurv>, ")
            print("               <j.B>, LPK")
        elif plot_name == "iota":
            ax.plot(self.data["nflux"], self.wout["iotaf"].values, **kwargs)
            ax.set_xlabel("Normalized Flux")
            ax.set_ylabel("iota")
            ax.set_title("Rotational Transform")
            # ax.set(xlabel='s',ylabel='iota',aspect='square')
        elif plot_name == "q":
            ax.plot(self.data["nflux"], 1.0 / self.wout["iotaf"].values, **kwargs)
            ax.set_xlabel("Normalized Flux")
            ax.set_ylabel("q")
            ax.set_title("Safety Factor")
        elif plot_name == "pressure":
            ax.plot(self.data["nflux"], self.wout["presf"].values / 1000, **kwargs)
            ax.set_xlabel("Normalized Flux")
            ax.set_ylabel("Pressure [kPa]")
            ax.set_title("Pressure Profile")
        elif plot_name == "<Buco>":
            ax.plot(self.data["nflux"], self.wout["buco"].values, **kwargs)
            ax.set_xlabel("Normalized Flux")
            ax.set_ylabel("<B^u> [T]")
            ax.set_title("Flux surface Averaged B^u")
        elif plot_name == "<Bvco>":
            ax.plot(self.data["nflux"], self.wout["bvco"].values, **kwargs)
            ax.set_xlabel("Normalized Flux")
            ax.set_ylabel("<B^v> [T]")
            ax.set_title("Flux surface Averaged B^v")
        elif plot_name == "<jcuru>":
            ax.plot(self.data["nflux"], self.wout["jcuru"].values / 1000, **kwargs)
            ax.set_xlabel("Normalized Flux")
            ax.set_ylabel("<j^u> [kA/m^2]")
            ax.set_title("Flux surface Averaged j^u")
        elif plot_name == "<jcurv>":
            ax.plot(self.data["nflux"], self.wout["jcurv"].values / 1000, **kwargs)
            ax.set_xlabel("Normalized Flux")
            ax.set_ylabel("<j^v> [kA/m^2]")
            ax.set_title("Flux surface Averaged j^v")
        elif plot_name == "<j.B>":
            ax.plot(self.data["nflux"], self.wout["jdotb"].values / 1000, **kwargs)
            ax.set_xlabel("Normalized Flux")
            ax.set_ylabel("<j.B> [T*kA/m^2]")
            ax.set_title("Flux surface Averaged j.B")
        elif plot_name == "LPK":
            self.surface[-1].plot(zeta=0, color="red", label=r"$\phi=0$", **kwargs)
            self.surface[-1].plot(
                zeta=0.5 * np.pi / self.data["nfp"], color="green", label=r"$\phi=0.25$", **kwargs
            )
            self.surface[-1].plot(
                zeta=np.pi / self.data["nfp"], color="blue", label=r"$\phi=0.5$", **kwargs
            )
            ax.set_title("LPK Plot")
        elif plot_name[0] == "-":
            print(plot_name)
        else:
            return ax



import numpy as np
from .misc import read_focus_boundary, write_focus_boundary


class FourSurf(object):
    """
    toroidal surface in Fourier representation
    R = \sum RBC cos(mu-nv) + RBS sin(mu-nv)
    Z = \sum ZBC cos(mu-nv) + ZBS sin(mu-nv)
    """

    def __init__(self, xm=[], xn=[], rbc=[], zbs=[], rbs=[], zbc=[]):
        """Initialization with Fourier harmonics.

        Parameters:
          xm -- list or numpy array, array of m index (default: [])
          xn -- list or numpy array, array of n index (default: [])
          rbc -- list or numpy array, array of radial cosine harmonics (default: [])
          zbs -- list or numpy array, array of z sine harmonics (default: [])
          rbs -- list or numpy array, array of radial sine harmonics (default: [])
          zbc -- list or numpy array, array of z cosine harmonics (default: [])

        """
        self.xm = np.atleast_1d(xm)
        self.xn = np.atleast_1d(xn)
        self.rbc = np.atleast_1d(rbc)
        self.rbs = np.atleast_1d(rbs)
        self.zbc = np.atleast_1d(zbc)
        self.zbs = np.atleast_1d(zbs)
        self.mn = len(self.xn)
        return

    @classmethod
    def read_focus_input(cls, filename, Mpol=9999, Ntor=9999):
        """initialize surface from the FOCUS format input file 'plasma.boundary'

        Parameters:
          filename -- string, path + name to the FOCUS input boundary file
          Mpol -- maximum truncated poloidal mode number (default: 9999)
          Ntol -- maximum truncated toroidal mode number (default: 9999)

        Returns:
          fourier_surface class
        """
        focus = read_focus_boundary(filename)
        nfp = focus["nfp"]
        xm = focus["surface"]["xm"]
        xn = focus["surface"]["xn"]
        cond = np.logical_and(np.abs(xm) <= Mpol, np.abs(xn) <= Ntor)
        xm = xm[cond]
        xn = xn[cond]
        rbc = focus["surface"]["rbc"][cond]
        rbs = focus["surface"]["rbs"][cond]
        zbc = focus["surface"]["zbc"][cond]
        zbs = focus["surface"]["zbs"][cond]
        return cls(
            xm=np.array(xm),
            xn=np.array(xn) * nfp,
            rbc=np.array(rbc),
            rbs=np.array(rbs),
            zbc=np.array(zbc),
            zbs=np.array(zbs),
        )

    @classmethod
    def read_spec_input(cls, filename, Mpol=9999, Ntor=9999):
        """initialize surface from the SPEC input file '*.sp'

        Parameters:
          filename -- string, path + name to the FOCUS input boundary file
          Mpol -- maximum truncated poloidal mode number (default: 9999)
          Ntol -- maximum truncated toroidal mode number (default: 9999)

        Returns:
          fourier_surface class
        """
        import f90nml
        from misc import vmecMN

        spec = f90nml(filename)
        # spec['physicslist'] =
        Mpol = min(Mpol, spec["physicslist"]["MPOL"])
        Ntor = min(Ntor, spec["physicslist"]["NTOR"])
        xm, xn = vmecMN(Mpol, Ntor)
        return

    @classmethod
    def read_spec_output(cls, spec_out, ns=-1):
        """initialize surface from the ns-th interface SPEC output

        Parameters:
          spec_out -- SPEC class, SPEC hdf5 results
          ns -- integer, the index of SPEC interface (default: -1)

        Returns:
          fourier_surface class
        """
        # check if spec_out is in correct format
        # if not isinstance(spec_out, SPEC):
        #    raise TypeError("Invalid type of input data, should be SPEC type.")
        # get required data
        xm = spec_out.output.im
        xn = spec_out.output.in_
        rbc = spec_out.output.Rbc[ns, :]
        zbs = spec_out.output.Zbs[ns, :]
        if spec_out.input.physics.Istellsym:
            # stellarator symmetry enforced
            rbs = np.zeros_like(rbc)
            zbc = np.zeros_like(rbc)
        else:
            rbs = spec_out.output.Rbs[ns, :]
            zbc = spec_out.output.Zbc[ns, :]
        return cls(xm=xm, xn=xn, rbc=rbc, rbs=rbs, zbc=zbc, zbs=zbs)

    @classmethod
    def read_vmec_input(cls, filename, tol=1e-8):
        """initialize surface from the ns-th interface SPEC output

        Parameters:
          woutfile -- string, path + name to the wout file from VMEC output
          ns -- integer, the index of VMEC nested flux surfaces (default: -1)

        Returns:
          fourier_surface class
        """
        import f90nml

        nml = f90nml.read(filename)
        indata = nml["indata"]
        mpol = indata["mpol"]
        ntor = indata["ntor"]
        nfp = indata["nfp"]
        arr_rbc = np.array(indata["rbc"])
        arr_zbs = np.array(indata["zbs"])
        arr_rbc[arr_rbc == None] = 0
        arr_zbs[arr_zbs == None] = 0
        try:
            arr_rbs = np.array(indata["rbs"])
            arr_zbc = np.array(indata["zbc"])
            arr_rbs[arr_rbs == None] = 0
            arr_zbc[arr_zbc == None] = 0
        except KeyError:
            arr_rbs = np.zeros_like(arr_rbc)
            arr_zbc = np.zeros_like(arr_rbc)
        nmin, mmin = indata.start_index["rbc"]
        mlen, nlen = np.shape(indata["rbc"])
        xm = []
        xn = []
        rbc = []
        zbs = []
        rbs = []
        zbc = []
        for i in range(mlen):
            m = i + mmin
            if m > mpol:
                continue
            for j in range(nlen):
                n = j + nmin
                if n > ntor:
                    continue
                if (
                    abs(arr_rbc[i, j])
                    + abs(arr_zbs[i, j])
                    + abs(arr_rbs[i, j])
                    + abs(arr_zbc[i, j])
                    < tol
                ):
                    continue
                xm.append(m)
                xn.append(n * nfp)
                rbc.append(arr_rbc[i, j])
                zbs.append(arr_zbs[i, j])
                rbs.append(arr_rbs[i, j])
                zbc.append(arr_zbc[i, j])
        return cls(xm=xm, xn=xn, rbc=rbc, rbs=rbs, zbc=zbc, zbs=zbs)

    @classmethod
    def read_vmec_output(cls, woutfile, ns=-1):
        """initialize surface from the ns-th interface SPEC output

        Parameters:
          woutfile -- string, path + name to the wout file from VMEC output
          ns -- integer, the index of VMEC nested flux surfaces (default: -1)

        Returns:
          fourier_surface class
        """
        import xarray as ncdata  # read netcdf file

        vmec = ncdata.open_dataset(woutfile)
        xm = vmec["xm"].values
        xn = vmec["xn"].values
        rmnc = vmec["rmnc"].values
        zmns = vmec["zmns"].values
        rbc = rmnc[ns, :]
        zbs = zmns[ns, :]

        if vmec["lasym__logical__"].values:
            # stellarator symmetry enforced
            zmnc = vmec["zmnc"].values
            rmns = vmec["rmns"].values
            rbs = rmns[ns, :]
            zbc = zmnc[ns, :]
        else:
            rbs = np.zeros_like(rbc)
            zbc = np.zeros_like(rbc)
        return cls(xm=xm, xn=xn, rbc=rbc, rbs=rbs, zbc=zbc, zbs=zbs)

    @classmethod
    def read_winding_surfce(cls, filename, Mpol=9999, Ntor=9999):
        """initialize surface from the NESCOIL format input file 'nescin.xxx'

        Parameters:
          filename -- string, path + name to the NESCOIL input boundary file
          Mpol -- maximum truncated poloidal mode number (default: 9999)
          Ntol -- maximum truncated toroidal mode number (default: 9999)

        Returns:
          fourier_surface class
        """
        with open(filename, "r") as f:
            line = ""
            while "phip_edge" not in line:
                line = f.readline()
            line = f.readline()
            nfp = int(line.split()[0])
            # print "nfp:",nfp

            line = ""
            while "Current Surface" not in line:
                line = f.readline()
            line = f.readline()
            line = f.readline()
            # print "Number of Fourier modes in coil surface from nescin file: ",line
            num = int(line)
            xm = []
            xn = []
            rbc = []
            rbs = []
            zbc = []
            zbs = []
            line = f.readline()  # skip one line
            line = f.readline()  # skip one line
            for i in range(num):
                line = f.readline()
                line_list = line.split()
                m = int(line_list[0])
                n = int(line_list[1])
                if abs(m) > Mpol or abs(n) > Ntor:
                    continue
                xm.append(m)
                xn.append(n)
                rbc.append(float(line_list[2]))
                zbs.append(float(line_list[3]))
                rbs.append(float(line_list[4]))
                zbc.append(float(line_list[5]))
            # NESCOIL uses mu+nv, minus sign is added
            return cls(
                xm=np.array(xm),
                xn=-np.array(xn) * nfp,
                rbc=np.array(rbc),
                rbs=np.array(rbs),
                zbc=np.array(zbc),
                zbs=np.array(zbs),
            )

    def rz(self, theta, zeta, normal=False):
        """get r,z position of list of (theta, zeta)

        Parameters:
          theta -- float array_like, poloidal angle
          zeta -- float array_like, toroidal angle value
          normal -- logical, calculate the normal vector or not (default: False)

        Returns:
           r, z -- float array_like
           r, z, [rt, zt], [rz, zz] -- if normal
        """
        assert len(np.atleast_1d(theta)) == len(
            np.atleast_1d(zeta)
        ), "theta, zeta should be equal size"
        # mt - nz (in matrix)
        _mtnz = np.matmul(
            np.reshape(self.xm, (-1, 1)), np.reshape(theta, (1, -1))
        ) - np.matmul(np.reshape(self.xn, (-1, 1)), np.reshape(zeta, (1, -1)))
        _cos = np.cos(_mtnz)
        _sin = np.sin(_mtnz)

        r = np.matmul(np.reshape(self.rbc, (1, -1)), _cos) + np.matmul(
            np.reshape(self.rbs, (1, -1)), _sin
        )
        z = np.matmul(np.reshape(self.zbc, (1, -1)), _cos) + np.matmul(
            np.reshape(self.zbs, (1, -1)), _sin
        )

        if not normal:
            return (r.ravel(), z.ravel())
        else:
            rt = np.matmul(np.reshape(self.xm * self.rbc, (1, -1)), -_sin) + np.matmul(
                np.reshape(self.xm * self.rbs, (1, -1)), _cos
            )
            zt = np.matmul(np.reshape(self.xm * self.zbc, (1, -1)), -_sin) + np.matmul(
                np.reshape(self.xm * self.zbs, (1, -1)), _cos
            )

            rz = np.matmul(np.reshape(-self.xn * self.rbc, (1, -1)), -_sin) + np.matmul(
                np.reshape(-self.xn * self.rbs, (1, -1)), _cos
            )
            zz = np.matmul(np.reshape(-self.xn * self.zbc, (1, -1)), -_sin) + np.matmul(
                np.reshape(-self.xn * self.zbs, (1, -1)), _cos
            )
            return (
                r.ravel(),
                z.ravel(),
                [rt.ravel(), zt.ravel()],
                [rz.ravel(), zz.ravel()],
            )

    def xyz(self, theta, zeta, normal=False):
        """get x,y,z position of list of (theta, zeta)

        Parameters:
          theta -- float array_like, poloidal angle
          zeta -- float array_like, toroidal angle value
          normal -- logical, calculate the normal vector or not (default: False)

        Returns:
           x, y, z -- float array_like
           x, y, z, [nx, ny, nz] -- if normal
        """
        data = self.rz(theta, zeta, normal)
        r = data[0]
        z = data[1]
        _sin = np.sin(np.ravel(zeta))
        _cos = np.cos(np.ravel(zeta))
        if not normal:
            return (r * _cos, r * _sin, z)
        else:
            _xt = data[2][0] * _cos  # dx/dtheta
            _yt = data[2][0] * _sin  # dy/dtheta
            _zt = data[2][1]  # dz/dtheta
            _xz = data[3][0] * _cos - r * _sin  # dx/dzeta
            _yz = data[3][0] * _sin + r * _cos  # dy/dzeta
            _zz = data[3][1]  # dz/dzeta
            # n = dr/dz x  dr/dt
            n = np.cross(np.transpose([_xz, _yz, _zz]), np.transpose([_xt, _yt, _zt]))
            return (r * _cos, r * _sin, z, n)

    def _areaVolume(
        self,
        theta0=0.0,
        theta1=2 * np.pi,
        zeta0=0.0,
        zeta1=2 * np.pi,
        npol=360,
        ntor=360,
    ):
        """Internel function to get surface area and volume

        Parameters:
          theta0 -- float, starting poloidal angle (default: 0.0)
          theta1 -- float, ending poloidal angle (default: 2*np.pi)
          zeta0 -- float, starting toroidal angle (default: 0.0)
          zeta1 -- float, ending toroidal angle (default: 2*np.pi)
          npol -- integer, number of poloidal discretization points (default: 360)
          ntor -- integer, number of toroidal discretization points (default: 360)

        Returns:
          area -- surface area
          volume -- surface volume
        """
        # get mesh data
        _theta = np.linspace(theta0, theta1, npol, endpoint=False)
        _zeta = np.linspace(zeta0, zeta1, ntor, endpoint=False)
        _tv, _zv = np.meshgrid(_theta, _zeta, indexing="ij")
        _x, _y, _z, _n = self.xyz(_tv, _zv, normal=True)
        # calculates the area and volume
        _dt = (theta1 - theta0) / npol
        _dz = (zeta1 - zeta0) / ntor
        _nn = np.linalg.norm(_n, axis=1)
        area = np.sum(_nn) * _dt * _dz
        volume = abs(np.sum(_x * _n[:, 0])) * _dt * _dz
        return area, volume

    def get_area(self):
        """Get the surface area and saved in self.area
        More comprehensive options can be found in self._areaVolume()

        Parameters:
           None

        Returns:
           area
        """
        self.area, _volume = self._areaVolume()
        return self.area

    def get_volume(self):
        """Get the surface volume and saved in self.volume
        More comprehensive options can be found in self._areaVolume()

        Parameters:
           None

        Returns:
           volume
        """
        _area, self.volume = self._areaVolume()
        return self.volume

    def plot(self, zeta=0.0, npoints=360, **kwargs):
        """plot the cross-section at zeta using matplotlib.pyplot

        Parameters:
          zeta -- float, toroidal angle value
          npoints -- integer, number of discretization points (default: 360)
          kwargs -- optional keyword arguments for pyplot

        Returns:
           line class in matplotlib.pyplot
        """
        import matplotlib.pyplot as plt

        # get figure and ax data
        if plt.get_fignums():
            fig = plt.gcf()
            ax = plt.gca()
        else:
            fig, ax = plt.subplots()
        # set default plotting parameters
        if kwargs.get("linewidth") == None:
            kwargs.update({"linewidth": 2.0})  # prefer thicker lines
        if kwargs.get("label") == None:
            kwargs.update({"label": "toroidal surface"})  # default label
        # get (r,z) data
        _r, _z = self.rz(np.linspace(0, 2 * np.pi, npoints), zeta * np.ones(npoints))
        line = ax.plot(_r, _z, **kwargs)
        plt.axis("equal")
        plt.xlabel("R [m]", fontsize=20)
        plt.ylabel("Z [m]", fontsize=20)
        plt.xticks(fontsize=16)
        plt.yticks(fontsize=16)
        return line

    def plot3d(
        self,
        engine="pyplot",
        theta0=0.0,
        theta1=2 * np.pi,
        zeta0=0.0,
        zeta1=2 * np.pi,
        npol=360,
        ntor=360,
        normal=False,
        fig=None,
        ax=None,
        show=True,
        **kwargs
    ):
        """plot 3D shape of the surface

        Parameters:
          engine -- string, plotting engine {'pyplot' (default), 'mayavi', 'plotly', 'noplot'}
          theta0 -- float, starting poloidal angle (default: 0.0)
          theta1 -- float, ending poloidal angle (default: 2*np.pi)
          zeta0 -- float, starting toroidal angle (default: 0.0)
          zeta1 -- float, ending toroidal angle (default: 2*np.pi)
          npol -- integer, number of poloidal discretization points (default: 360)
          ntor -- integer, number of toroidal discretization points (default: 360)
          normal -- bool, if calculating the normal vector (default: False)
          fig -- , figure to be plotted on (default: None)
          ax -- , axis to be plotted on (default: None)
          show -- bool, if show the plotly figure immediately (default: True)
          kwargs -- optional keyword arguments for plotting

        Returns:
           xsurf, ysurf, zsurf -- arrays of x,y,z coordinates on the surface
        """
        # get mesh data
        _theta = np.linspace(theta0, theta1, npol)
        _zeta = np.linspace(zeta0, zeta1, ntor)
        _tv, _zv = np.meshgrid(_theta, _zeta, indexing="ij")
        if normal:
            _x, _y, _z, _n = self.xyz(_tv, _zv, normal=normal)
            n = np.reshape(_n, (3, npol, ntor))
        else:
            _x, _y, _z = self.xyz(_tv, _zv)
            n = None
        xsurf = np.reshape(_x, (npol, ntor))
        ysurf = np.reshape(_y, (npol, ntor))
        zsurf = np.reshape(_z, (npol, ntor))
        if engine == "noplot":
            # just return xyz data
            pass
        elif engine == "pyplot":
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D

            # plot in matplotlib.pyplot
            if ax is None or ax.name != "3d":
                fig = plt.figure()
                ax = fig.add_subplot(111, projection="3d")
            ax.plot_surface(xsurf, ysurf, zsurf, **kwargs)
        elif engine == "mayavi":
            # plot 3D surface in mayavi.mlab
            from mayavi import mlab  # to overrid plt.mlab

            mlab.mesh(xsurf, ysurf, zsurf, **kwargs)
        elif engine == "plotly":
            import plotly.graph_objects as go

            if "color" in list(kwargs.keys()):
                color = kwargs["color"]
                del kwargs["color"]
                kwargs["colorscale"] = [[0, color], [1, color]]
            if fig is None:
                fig = go.Figure()
            fig.add_trace(go.Surface(x=xsurf, y=ysurf, z=zsurf, **kwargs))
            fig.update_layout(scene_aspectmode="data")
            if show:
                fig.show()
        else:
            raise ValueError("Invalid engine option {pyplot, mayavi, noplot}")
        return (xsurf, ysurf, zsurf, n)

    def toVTK(self, vtkname, npol=360, ntor=360, **kwargs):
        """save surface shape a vtk grid file

        Parameters:
          vtkname -- string, the filename you want to save, final name is 'vtkname.vts'
          npol -- integer, number of poloidal discretization points (default: 360)
          ntor -- integer, number of toroidal discretization points (default: 360)
          kwargs -- optional keyword arguments for saving as pointdata

        Returns:

        """
        from pyevtk.hl import gridToVTK  # save to binary vtk

        _xx, _yy, _zz, _nn = self.plot3d(
            "noplot",
            zeta0=0.0,
            zeta1=2 * np.pi,
            theta0=0.0,
            theta1=2 * np.pi,
            npol=npol,
            ntor=ntor,
            normal=True,
        )
        _xx = _xx.reshape((1, npol, ntor))
        _yy = _yy.reshape((1, npol, ntor))
        _zz = _zz.reshape((1, npol, ntor))
        _nn = _nn.reshape((1, npol, ntor, 3))

        kwargs.setdefault(
            "n",
            (
                np.ascontiguousarray(_nn[:, :, :, 0]),
                np.ascontiguousarray(_nn[:, :, :, 1]),
                np.ascontiguousarray(_nn[:, :, :, 2]),
            ),
        )
        gridToVTK(vtkname, _xx, _yy, _zz, pointData=kwargs)
        return

    def toSTL(self, stlname, **kwargs):
        """save surface shape a stl file using meshio

        Parameters:
          stlname -- string, the filename you want to save, final name is 'stlname.vts'
          kwargs -- optional keyword arguments used for self.plot3d.

        Returns:
          mesh: Mesh object in meshio

        """
        import meshio

        kwargs.setdefault("npol", 120)
        kwargs.setdefault("ntor", 180)
        _xx, _yy, _zz, _nn = self.plot3d("noplot", **kwargs)
        npol = kwargs["npol"]
        ntor = kwargs["ntor"]
        points = np.ascontiguousarray(
            np.transpose([_xx.ravel(), _yy.ravel(), _zz.ravel()])
        )
        con = []
        for i in range(npol - 1):
            for j in range(ntor - 1):
                ij = i * ntor + j
                con.append([ij, ij + 1, ij + ntor + 1])
                con.append([ij, ij + ntor, ij + ntor + 1])
        cells = [("triangle", np.array(con))]
        mesh = meshio.Mesh(points, cells)
        mesh.write(stlname)
        return mesh

    def write_focus_input(self, filename, nfp=1, bn=None):
        """Write the Fourier harmonics down in FOCUS format

        Args:
            filename ([type]): Output file name.
            nfp (int, optional): Number of toroidal periodicity Defaults to 1.
            bn (dict, optional): Bn dict, containing 'xm', 'xn', 'bnc', 'bns'. Defaults to None.
        """
        surf = {}
        surf["xn"] = self.xn // nfp
        surf["xm"] = self.xm
        surf["rbc"] = self.rbc
        surf["rbs"] = self.rbs
        surf["zbc"] = self.zbc
        surf["zbs"] = self.zbs
        write_focus_boundary(filename, surf, nfp, bn)
        return

    def write_vmec_input(self, filename, template=None, nfp=1, **kwargs):
        import f90nml

        if template is None:
            with open(filename, "w") as f:
                for i in range(self.mn):
                    f.write(
                        "RBC({n:d},{m:d}) = {r:15.7E} \t ZBS({n:d},{m:d}) = {z:15.7E} \n".format(
                            n=int(self.xn[i]) // nfp,
                            m=int(self.xm[i]),
                            r=self.rbc[i],
                            z=self.zbs[i],
                        )
                    )
        else:
            pass
        return

    def grid_box(self, ntor=64, npol=64):
        """Return the max R & Z values of the surface

        Args:
            ntor (int, optional): Toroidal resolution. Defaults to 64.
            npol (int, optional): Poloidal resolution. Defaults to 64.

        Returns:
            (Rmin, Rmax, Zmin, Zmax): the max R & Z values
        """
        _theta = np.linspace(0, 2 * np.pi, npol)
        _zeta = np.linspace(0, 2 * np.pi, ntor)
        _tv, _zv = np.meshgrid(_theta, _zeta, indexing="ij")
        data = self.rz(_tv, _zv)
        r = data[0]
        z = data[1]
        return (np.min(r), np.max(r), np.min(z), np.max(z))

    def __del__(self):
        class_name = self.__class__.__name__
